#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#include <stdlib.h>

#ifndef NO_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#ifndef __WIN32__
#include <unistd.h>
#endif

#include "timidity.h"
#include "common.h"
#include "instrum.h"
#include "playmidi.h"
#include "readmidi.h"
#include "controls.h"
#include "output.h"
#include "arc.h"
#include "wrd.h"

#define USER_EXCLUSIVE_LENGTH 24
#define MAX_EXCLUSIVE_LENGTH 1024
static uint8 user_exclusive_data[8][USER_EXCLUSIVE_LENGTH];
static int32 init_tempo;
static int play_bias;


/*ARGSUSED*/
static int null_cmsg(int type, int verbosity_level, char *fmt, ...){return 0;}
static void null_proc(){}
static ControlMode null_control_mode =
{
    "", 0,
    1,0,0,
    0,
    NULL, null_proc, NULL, NULL, null_cmsg, null_proc
};

static PlayMode null_play_mode =
{
    8000, PE_MONO, 0,
    -1,
    {0,0,0,0,0},
    "", 0,
    "",
    NULL, NULL, NULL, NULL
};

/*ARGSUSED*/
static int  null_wrdt_open(char *wrdt_opts) { return 0; }
/*ARGSUSED*/
static void null_wrdt_apply(int cmd, int argc, int args[]) { }
static void null_wrdt_update_events(void) { }
static void null_wrdt_end(void) { }
static void null_wrdt_close(void) { }
static WRDTracer null_wrdt_mode =
{
    "No WRD trace", '-',
    0,
    null_wrdt_open,
    null_wrdt_apply,
    NULL,
    null_wrdt_update_events,
    NULL,
    null_wrdt_end,
    null_wrdt_close
};

PlayMode *play_mode = &null_play_mode;
ControlMode *ctl = &null_control_mode;
WRDTracer *wrdt = &null_wrdt_mode;

static struct
{
    char *name;
    char *info;
} control_names[] =
{
    {"Bank select MSB",		"0-127, MSB"},	/* 0 */
    {"Modulation wheel MSB",	"0-127, MSB"},	/* 1 */
    {"Breath control MSB",	"0-127, MSB"},	/* 2 */
    {NULL,			"0-127, MSB"},	/* 3 */
    {"Foot controller MSB",	"0-127, MSB"},	/* 4 */
    {"Portamento time MSB",	"0-127, MSB"},	/* 5 */
    {"Data entry MSB",		"0-127, MSB"},	/* 6 */
    {"Main volume MSB",		"0-127, MSB"},	/* 7 */
    {"Balance MSB",		"0-127, MSB"},	/* 8 */
    {NULL,			"0-127, MSB"},	/* 9 */
    {"Pan position MSB",	"0-127, MSB"},	/* 10 */
    {"Expression MSB",		"0-127, MSB"},	/* 11 */
    {"Effect control 1 MSB",	"0-127, MSB"},	/* 12 */
    {"Effect control 2/control change MSB","0-127"},/* 13 */
    {NULL,			"0-127, MSB"},	/* 14 */
    {NULL,			"0-127, MSB"},	/* 15 */
    {"General purpose control 1 MSB","0-127, MSB"},	/* 16 */
    {"General purpose control 2 MSB","0-127, MSB"},	/* 17 */
    {"General purpose control 3 MSB","0-127, MSB"},	/* 18 */
    {"General purpose control 4 MSB","0-127, MSB"},	/* 19 */
    {NULL, "0-127, MSB"},/* 20 */
    {NULL, "0-127, MSB"},/* 21 */
    {NULL, "0-127, MSB"},/* 22 */
    {NULL, "0-127, MSB"},/* 23 */
    {NULL, "0-127, MSB"},/* 24 */
    {NULL, "0-127, MSB"},/* 25 */
    {NULL, "0-127, MSB"},/* 26 */
    {NULL, "0-127, MSB"},/* 27 */
    {NULL, "0-127, MSB"},/* 28 */
    {NULL, "0-127, MSB"},/* 29 */
    {NULL, "0-127, MSB"},/* 30 */
    {NULL, "0-127, MSB"},/* 31 */
    {"Bank select LSB",		"0-127, LSB"},	/* 32 */
    {"Modulation wheel LSB",	"0-127, LSB"},	/* 33 */
    {"Breath control LSB",	"0-127, LSB"},	/* 34 */
    {NULL,			"0-127, LSB"},	/* 35 */
    {"Foot controller LSB",	"0-127, LSB"},	/* 36 */
    {"Portamento time LSB",	"0-127, LSB"},	/* 37 */
    {"Data entry LSB",		"0-127, LSB"},	/* 38 */
    {"Main volume LSB",		"0-127, LSB"},	/* 39 */
    {"Balance LSB",		"0-127, LSB"},	/* 40 */
    {"Pan position LSB",	"0-127, LSB"},	/* 41 */
    {"Expression LSB",		"0-127, LSB"},	/* 42 */
    {"Effect control 1 LSB",	"0-127, LSB"},	/* 43 */
    {"Effect control 2 LSB",	"0-127, LSB"},	/* 44 */
    {NULL, "0-127, LSB"},/* 45 */
    {NULL, "0-127, LSB"},/* 46 */
    {NULL, "0-127, LSB"},/* 47 */
    {NULL, "0-127, LSB"},/* 48 */
    {NULL, "0-127, LSB"},/* 49 */
    {NULL, "0-127, LSB"},/* 50 */
    {NULL, "0-127, LSB"},/* 51 */
    {NULL, "0-127, LSB"},/* 52 */
    {NULL, "0-127, LSB"},/* 53 */
    {NULL, "0-127, LSB"},/* 54 */
    {NULL, "0-127, LSB"},/* 55 */
    {NULL, "0-127, LSB"},/* 56 */
    {NULL, "0-127, LSB"},/* 57 */
    {NULL, "0-127, LSB"},/* 58 */
    {NULL, "0-127, LSB"},/* 59 */
    {NULL, "0-127, LSB"},/* 60 */
    {NULL, "0-127, LSB"},/* 61 */
    {NULL, "0-127, LSB"},/* 62 */
    {NULL, "0-127, LSB"},/* 63 */
    {"Damper pedal on/off (Sustain)",	"0=off, 127=on"},/* 64 */
    {"Portamento on/off",	"0=off, 127=on"},/* 65 */
    {"Sustenuto on/off",	"0=off, 127=on"},/* 66 */
    {"Soft pedal on/off",	"0=off, 127=on"},/* 67 */
    {"Legato pedal on/off",	"0=off, 127=on"},/* 68 */
    {"Hold 2 pedal on/off",	"0=off, 127=on"},/* 69 */
    {"Sound variation",		"0-127"},	/* 70 */
    {"Harmonic content/Sound timbre","0-127"},	/* 71 */
    {"Release time",		"0-127"},	/* 72 */
    {"Attack time",		"0-127"},	/* 73 */
    {"Brightness",		"0-127"},	/* 74 */
    {NULL,"0-127"},	/* 75 */
    {NULL,"0-127"},	/* 76 */
    {NULL,"0-127"},	/* 77 */
    {NULL,"0-127"},	/* 78 */
    {NULL,"0-127"},	/* 79 */
    {"General purpose control 5 (Mute)", NULL},/* 80 */
    {"General purpose control 6", NULL},/* 81 */
    {"General purpose control 7", NULL},/* 82 */
    {"General purpose control 8(Portamento control (64 = C4))", NULL},/* 83 */
    {NULL, NULL},	/* 84 */
    {NULL, NULL},	/* 85 */
    {NULL, NULL},	/* 86 */
    {NULL, NULL},	/* 87 */
    {NULL, NULL},	/* 88 */
    {NULL, NULL},	/* 89 */
    {NULL, NULL},	/* 90 */
    {"Reverb",			"0-127"},	/* 91 */
    {"Tremulo depth",		"0-127"},	/* 92 */
    {"Chorus depth",		"0-127"},	/* 93 */
    {"Celeste (detune) depth",	"0-127"},	/* 94 */
    {"Phaser level",		"0-127"},	/* 95 */
    {"Data entry +1",		"0"},		/* 96 */
    {"Data entry -1",		"0"},		/* 97 */
    {"NRPN LSB", NULL},	/* 98 */
    {"NRPN MSB", NULL},	/* 99 */
    {"RPN LSB",	NULL},	/* 100 */
    {"RPN MSB", NULL},	/* 101 */
    {NULL, NULL},	/* 102 */
    {NULL, NULL},	/* 103 */
    {NULL, NULL},	/* 104 */
    {NULL, NULL},	/* 105 */
    {NULL, NULL},	/* 106 */
    {NULL, NULL},	/* 107 */
    {NULL, NULL},	/* 108 */
    {NULL, NULL},	/* 109 */
    {NULL, NULL},	/* 110 */
    {NULL, NULL},	/* 111 */
    {NULL, NULL},	/* 112 */
    {NULL, NULL},	/* 113 */
    {NULL, NULL},	/* 114 */
    {NULL, NULL},	/* 115 */
    {NULL, NULL},	/* 116 */
    {NULL, NULL},	/* 117 */
    {NULL, NULL},	/* 118 */
    {NULL, NULL},	/* 119 */
    {"All sound off",		"0"},			/* 120 */
    {"All controllers off",	"0"},			/* 121 */
    {"Local control on/off",	"0=off 127=on"},	/* 122 */
    {"All notes off",		"0"},			/* 123 */
    {"Omni mode off (includes all notes off)",	"0"},	/* 124 */
    {"Omni mode on (includes all notes off)",	"0"},	/* 125 */
    {"Poly mode on/off(includes all notes off)", ""},	/* 126 */
    {"Poly mode on(incl mono=off&all notes off)","0"},	/* 127 */
};

static void rcpdump(char *name);
static int rcptrack(struct timidity_file *tf, int type);
static char *rcp_cmd_name(int cmd, int a);

extern struct URL_module URL_module_file;
extern struct URL_module URL_module_dir;
#ifdef SUPPORT_SOCKET
extern struct URL_module URL_module_http;
extern struct URL_module URL_module_ftp;
extern struct URL_module URL_module_news;
extern struct URL_module URL_module_newsgroup;
#endif /* SUPPORT_SOCKET */

int main(int argc, char **argv)
{
    int i;
    int nfiles;
    char **files;

    url_add_modules(
	&URL_module_file,
	&URL_module_dir,
#ifdef SUPPORT_SOCKET
	&URL_module_http,
	&URL_module_ftp,
	&URL_module_news,
	&URL_module_newsgroup,
#endif /* SUPPORT_SOCKET */
	NULL);

    if(argc == 1)
    {
	rcpdump("-");
	return 0;
    }

    nfiles = argc - 1;
    files  = argv + 1;

    files = expand_file_archives(files, &nfiles);
    for(i = 0; i < nfiles; i++)
	rcpdump(files[i]);
    free_archive_files();
    return 0;
}

static void rcpdump(char *fname)
{
    struct timidity_file *tf;
    char buff[361], *p;
    int ntrack, timebase1, timebase2, i, j, len;
    int file_type;

    printf("%s:\n", fname);

    if((tf = open_file(fname, 1, OF_SILENT)) == NULL)
    {
	perror(fname);
	return;
    }

    if(!IS_URL_SEEK_SAFE(tf->url))
	if((tf->url = url_cache_open(tf->url, 1)) == NULL)
	{
	    close_file(tf);
	    return;
	}

    if(tf_read(buff, 1, 32, tf) != 32)
    {
	close_file(tf);
	return; /* Error */
    }

    if(strncmp(buff, "RCM-PC98V2.0(C)COME ON MUSIC", 28) == 0)
	file_type = IS_RCP_FILE; /* RCP or R36 */
    else if(strncmp(buff, "COME ON MUSIC RECOMPOSER RCP3.0", 31) == 0)
	file_type = IS_G36_FILE; /* G36 or G18 */
    else
    {
	close_file(tf);
	return; /* Error */
    }

    
    /* title */
    if(tf_read(buff, 1, 64, tf) != 64)
    {
	close_file(tf);
	return;
    }
    buff[64] = '\0';
    for(len = 63; len >= 0; len--)
    {
	if(buff[len] == ' ')
	    buff[len] = '\0';
	else if(buff[len] != '\0')
	    break;
    }
    
    len = SAFE_CONVERT_LENGTH(len + 1);
    p = (char *)new_segment(&tmpbuffer, len);
    code_convert(buff, p, len, NULL, NULL);
    printf("Title: %s" NLS, p);
    reuse_mblock(&tmpbuffer);

    if(file_type != IS_G36_FILE) /* RCP or R36 */
    {
	if(tf_read(buff, 1, 336, tf) != 336)
	{
	    close_file(tf);
	    return;
	}
	buff[336] = '\0';
	for(len = 335; len >= 0; len--)
	{
	    if(buff[len] == ' ')
		buff[len] = '\0';
	    else if(buff[len] != '\0')
		break;
	}
	len = SAFE_CONVERT_LENGTH(len + 1);
	p = (char *)new_segment(&tmpbuffer, len);
	code_convert(buff, p, len, NULL, NULL);
	printf("Memo: %s" NLS, p);
	reuse_mblock(&tmpbuffer);

	skip(tf, 16);		/* 0x40 */

	timebase1 = tf_getc(tf);
	init_tempo = tf_getc(tf); /* tempo */
	printf("Tempo: %ld" NLS, init_tempo);
	if(init_tempo < 8 || init_tempo > 250)
	    init_tempo = 120;

	i = tf_getc(tf);
	j = tf_getc(tf);
	printf("KeySignature: %d/%d" NLS, i, j);
	printf("TimeSignature: %d" NLS, tf_getc(tf));

	play_bias = (int)(signed char)tf_getc(tf);
	printf("PlayBias: %d" NLS, play_bias);
	if(play_bias < -36 || play_bias > 36)
	    play_bias = 0;

	tf_read(buff, 1, 12, tf);
	buff[12] = '\0';
	printf("CM6file: %s\n", buff);
	skip(tf, 4);		/* reserved */

	tf_read(buff, 1, 12, tf);
	buff[12] = '\0';
	printf("GSDfile: %s\n", buff);
	skip(tf, 4);		/* reserved */

	if((ntrack = tf_getc(tf)) == EOF)
	{
	    close_file(tf);
	    return;
	}
	printf("Tracks: %d\n", ntrack);
	if(ntrack != 18 && ntrack != 36)
	    ntrack = 18;
	timebase2 = tf_getc(tf);
	skip(tf, 14);		/* reserved */
	skip(tf, 16);		/*  */

	for(i = 0; i < 32; i++)
	{
	    int key, gate;

	    tf_read(buff, 1, 14, tf);
	    buff[14] = '\0';
	    len = SAFE_CONVERT_LENGTH(14);
	    p = (char *)new_segment(&tmpbuffer, len);
	    code_convert(buff, p, len, NULL, NULL);
	    key = tf_getc(tf);
	    gate = tf_getc(tf);
	    printf("Rhythm %d: %s key=%d gate=%d\n", i, p, key, gate);
	    reuse_mblock(&tmpbuffer);
	}
    }
    else /* G36 or G18 */
    {
	skip(tf, 64);	/* reserved */

	/* memo */
	if(tf_read(buff, 1, 360, tf) != 360)
	{
	    close_file(tf);
	    return;
	}
	buff[360] = '\0';
	for(len = 359; len >= 0; len--)
	{
	    if(buff[len] == ' ')
		buff[len] = '\0';
	    else if(buff[len] != '\0')
		break;
	}
	len = SAFE_CONVERT_LENGTH(len + 1);
	p = (char *)new_segment(&tmpbuffer, len);
	code_convert(buff, p, len, NULL, NULL);
	printf("Memo: %s" NLS, p);
	reuse_mblock(&tmpbuffer);

	/* Number of tracks */
	ntrack = tf_getc(tf);
	printf("Tracks: %d\n", ntrack);
	if(ntrack != 18 && ntrack != 36)
	    ntrack = 18;
	tf_getc(tf);
	timebase1 = tf_getc(tf);
	timebase2 = tf_getc(tf);
	init_tempo = tf_getc(tf); /* tempo */
	printf("Tempo: %ld" NLS, init_tempo);
	if(init_tempo < 8 || init_tempo > 250)
	    init_tempo = 120;
	tf_getc(tf); /* ##?? */

	i = tf_getc(tf);
	j = tf_getc(tf);
	printf("KeySignature: %d/%d" NLS, i, j);
	printf("TimeSignature: %d" NLS, tf_getc(tf));

	play_bias = (int)(signed char)tf_getc(tf);
	printf("PlayBias: %d" NLS, play_bias);
	if(play_bias < -36 || play_bias > 36)
	    play_bias = 0;
	skip(tf, 6 + 16 + 112);	/* reserved */

	tf_read(buff, 1, 12, tf);
	buff[12] = '\0';
	printf("GSDfile(1): %s\n", buff);
	skip(tf, 4);		/* reserved */

	tf_read(buff, 1, 12, tf);
	buff[12] = '\0';
	printf("GSDfile(2): %s\n", buff);
	skip(tf, 4);		/* reserved */

	tf_read(buff, 1, 12, tf);
	buff[12] = '\0';
	printf("CM6file: %s\n", buff);
	skip(tf, 4);		/* reserved */

	skip(tf, 80);		/* reserved */

	for(i = 0; i < 128; i++)
	{
	    int key, gate;

	    tf_read(buff, 1, 14, tf);
	    buff[14] = '\0';
	    len = SAFE_CONVERT_LENGTH(14);
	    p = (char *)new_segment(&tmpbuffer, len);
	    code_convert(buff, p, len, NULL, NULL);
	    key = tf_getc(tf);
	    gate = tf_getc(tf);
	    printf("Rhythm %d: %s key=%d gate=%d\n", i, p, key, gate);
	    reuse_mblock(&tmpbuffer);
	}
    }

    /* SysEx data */
    for(i = 0; i < 8; i++)
    {
	/* memo */
	tf_read(buff, 1, 24, tf);
	buff[24] = '\0';
	len = strlen(buff);
	len = SAFE_CONVERT_LENGTH(len);
	p = (char *)new_segment(&tmpbuffer, len);
	code_convert(buff, p, len, NULL, NULL);
	printf("UserSysEx %d: %s:", i, p);
	reuse_mblock(&tmpbuffer);

	if(tf_read(user_exclusive_data[i], 1, USER_EXCLUSIVE_LENGTH, tf)
	   != USER_EXCLUSIVE_LENGTH)
	{
	    close_file(tf);
	    printf(" EOF\n");
	    return;
	}
	for(j = 0; j < USER_EXCLUSIVE_LENGTH; j++)
	    printf(" %02x", user_exclusive_data[i][j]);
	printf("\n");
    }

    printf("Timebase: %d\n", (timebase1 | (timebase2 << 8)));

    for(i = 0; i < ntrack; i++)
	if(rcptrack(tf, file_type))
	    break; /* Error */
    close_file(tf);
}

static int rcptrack(struct timidity_file *tf, int file_type)
{
    int size;
    int ch;
    long last_point, last_point_offset;
    int key_offset;
    int i, len;

#if 0
    uint8 sysex[MAX_EXCLUSIVE_LENGTH];
    int32 current_tempo;
    int roland_base_init = 0;
    int roland_dev_init = 0;
    int roland_base_addr0 = 0;
    int roland_base_addr1 = 0;
    int roland_device_id = 0x10;
    int roland_model_id = 0x16;
    int yamaha_base_init = 0;
    int yamaha_dev_init = 0;
    int yamaha_base_addr0 = 0;
    int yamaha_base_addr1 = 0;
    int yamaha_device_id = 0x10;	/* ??? */
    int yamaha_model_id = 0x16;		/* ??? */
    unsigned char cs;
    long same_measure = 0;
#endif

    int trackno;
    long track_top;
    int time_offset;
    char buff[64], *p;
    int32 at;
    int rhythm;

    /*
     * Read Track Header
     */

    track_top = tf_tell(tf);
    size = tf_getc(tf);
    size |= (tf_getc(tf) << 8);
    printf("TrackSize %d\n", size);
    last_point = size + tf_tell(tf) - 2;

    if(file_type == IS_G36_FILE)
    {
	tf_getc(tf);
	tf_getc(tf);
	last_point_offset = 6;
    }
    else
	last_point_offset = 4;
    trackno = tf_getc(tf);		/* Track Number */
    printf("Track %d" NLS, trackno);
    rhythm = tf_getc(tf);	/* Rhythm */
    printf("Rhythm %d\n", rhythm);

    if((ch = tf_getc(tf)) == 0xff) /* No playing */
    {
	puts("No playing");
	tf_seek(tf, last_point, SEEK_SET);
	return 0;
    }

    printf("Channel %d" NLS, ch);

    /* Key offset */
    key_offset = tf_getc(tf);
    printf("KeyOffset %d" NLS, key_offset);

    /* Time offset */
    time_offset = (int)(signed char)tf_getc(tf);
    printf("TimeOffset %d" NLS, time_offset);

    time_offset += 99;
    i = tf_getc(tf);		/* mode */
    if(i == 1)
    {
	/* Mute */
	puts("Mute");
	tf_seek(tf, last_point, SEEK_SET);
	return 0;
    }

    /* Comment */
    tf_read(buff, 1, 36, tf);
    buff[36] = '\0';
    for(len = 35; len >= 0; len--)
    {
	if(buff[len] == ' ')
	    buff[len] = '\0';
	else if(buff[len] != '\0')
	    break;
    }
    len = SAFE_CONVERT_LENGTH(len);
    p = (char *)new_segment(&tmpbuffer, len);
    code_convert(buff, p, len, NULL, NULL);
    printf("TrackName %s" NLS, p);
    reuse_mblock(&tmpbuffer);

    /*
     * End of Track Header
     */

    at = 0;
    while(last_point >= tf_tell(tf) + last_point_offset)
    {
	uint8 cmds[6];
	int step;

	if(file_type != IS_G36_FILE)
	{
	    if(tf_read(cmds, 1, 4, tf) != 4)
		break;
	    i = tf_tell(tf) - 4 - track_top;
	    printf("[%d,0x%x] %ld %s: %02x %02x %02x %02x" NLS,
		   i, i, at, rcp_cmd_name(cmds[0], cmds[2]),
		   cmds[0], cmds[1], cmds[2], cmds[3]);
	    step = cmds[1];
	}
	else
	{
	    if(tf_read(cmds, 1, 6, tf) != 6)
		break;
	    i = tf_tell(tf) - 6 - track_top;
	    printf("[%d,0x%x] %ld %s: %02x %02x %02x %02x %02x %02x" NLS,
		   i, i, at, rcp_cmd_name(cmds[0], cmds[4]),
		   cmds[0], cmds[1], cmds[2], cmds[3], cmds[4], cmds[5]);
	    step = (int)cmds[2] | ((int)cmds[3] << 8);
	}

	if(cmds[0] < 0x80)		/* Note event */
	{
	    at += step;
	    continue;
	}

	/* command */
	switch(cmds[0])
	{
	  case 0x90:	/* User Exclusive 1 */
	  case 0x91:	/* User Exclusive 2 */
	  case 0x92:	/* User Exclusive 3 */
	  case 0x93:	/* User Exclusive 4 */
	  case 0x94:	/* User Exclusive 5 */
	  case 0x95:	/* User Exclusive 6 */
	  case 0x96:	/* User Exclusive 7 */
	  case 0x97:	/* User Exclusive 8 */
	    at += step;
	    break;

	  case 0x98:	/* Channel Exclusive */
	    at += step;
	    break;

	  case 0xf7:	/* 2nd Event */
	    break;

	  case 0xe1:	/* Bank PGM */
	    at += step;
	    break;

	  case 0xe2:	/* bank & program change */
	    at += step;
	    break;

	  case 0xe5:	/* key scan */
	    at += step;
	    break;

	  case 0xe6:	/* MIDI channel change */
	  case 0xe7:	/* tempo change */
	  case 0xea:	/* channel after touch (channel pressure) */
	  case 0xeb:	/* control change */
	  case 0xec:	/* program change */
	  case 0xed:	/* after touch polyphonic (polyphonic key pressure) */
	  case 0xee:	/* pitch bend */
	  case 0xdd:	/* Roland base */
	  case 0xdf:	/* Roland device */
	  case 0xde:	/* Roland para */
	  case 0xd0:	/* Yamaha base */
	  case 0xd1:	/* Yamaha para */
	  case 0xd3:	/* XG para */
	    at += step;

	  case 0xf5:	/* key change */
	    break;

	  case 0xf6:	/* comment */
	    break;

	  case 0xf8:	/* loop end */
	    break;

	  case 0xf9:	/* loop start */
	    break;

	  case 0xfc:	/* same measure */
	    break;

	  case 0xfd:	/* measure end */
	    break;
	  case 0xfe:	/* end of track */
	    break;

	  case 0xc0:	/* DX7 function */
	  case 0xc1:	/* DX parameter */
	  case 0xc2:	/* DX RERF */
	  case 0xc3:	/* TX function */
	  case 0xc5:	/* FB-01 P parameter */
	  case 0xc6:	/* FB-01 S System */
	  case 0xc7:	/* TX81Z V VCED */
	  case 0xc8:	/* TX81Z A ACED */
	  case 0xc9:	/* TX81Z P PCED */
	  case 0xca:	/* TX81Z S System */
	  case 0xcb:	/* TX81Z E EFFECT */
	  case 0xcc:	/* DX7-2 R REMOTE SW */
	  case 0xcd:	/* DX7-2 A ACED */
	  case 0xce:	/* DX7-2 P PCED */
	  case 0xcf:	/* TX802 P PCED */
	  case 0xdc:	/* MKS-7 */
	    at += step;
	    break;
	  default:
	    at += step;
	    break;
	}
    }

    /* All note off */
    tf_seek(tf, last_point, SEEK_SET);
    return 0;
}

static char *rcp_cmd_name(int cmd, int a)
{
    static char name[256];

    if(cmd < 0x80)
    {
	sprintf(name, "NoteOn %d", cmd);
	return name;
    }

    switch(cmd)
    {
      case 0x90: return "UserExclusive0";
      case 0x91: return "UserExclusive1";
      case 0x92: return "UserExclusive2";
      case 0x93: return "UserExclusive3";
      case 0x94: return "UserExclusive4";
      case 0x95: return "UserExclusive5";
      case 0x96: return "UserExclusive6";
      case 0x97: return "UserExclusive7";
      case 0x98: return "ChannelExclusive";
      case 0xd0: return "YamahaBase";
      case 0xd1: return "YamahaPara";
      case 0xd2: return "YamahaDevice";
      case 0xd3: return "XGPara";
      case 0xdc: return "MKS-7";
      case 0xdd: return "RolandBase";
      case 0xde: return "RolandPara";
      case 0xdf: return "RolandDevice";
      case 0xe1: return "BnkLPrg";
      case 0xe2: return "Bank&ProgCng";
      case 0xe5: return "KeyScan";
      case 0xe6: return "ChChange";
      case 0xe7: return "TempoChange";
      case 0xea: return "ChannelAfterTouch";

      case 0xeb:
	if(a < 0x7f && control_names[a].name != NULL)
	    sprintf(name, "ControlChange %s", control_names[a].name);
	else
	    sprintf(name, "ControlChange %d", a);
	return name;

      case 0xec: return "ProgChange";
      case 0xed: return "AfterTouch";
      case 0xee: return "PitchBend";
      case 0xf5: return "KeyChange";
      case 0xf6: return "Comment";
      case 0xf7: return "2ndEvent";
      case 0xf8: return "LoopEnd";
      case 0xf9: return "LoopStart";
      case 0xfc: return "SameMeasure";
      case 0xfd: return "MeasureEnd";
      case 0xfe: return "EndOfTrack";
      case 0xc0: return "DX7 function";
      case 0xc1: return "DX parameter";
      case 0xc2: return "DX RERF";
      case 0xc3: return "TX function";
      case 0xc5: return "FB-01 P parameter";
      case 0xc6: return "FB-01 S System";
      case 0xc7: return "TX81Z V VCED";
      case 0xc8: return "TX81Z A ACED";
      case 0xc9: return "TX81Z P PCED";
      case 0xca: return "TX81Z S System";
      case 0xcb: return "TX81Z E EFFECT";
      case 0xcc: return "DX7-2 R REMOTE SW";
      case 0xcd: return "DX7-2 A ACED";
      case 0xce: return "DX7-2 P PCED";
      case 0xcf: return "TX802 P PCED";
    }

    sprintf(name, "Unknown-0x%02x", cmd);
    return name;
}
